/*
Copyright (C) 2002  Ulric Eriksson <ulric@siag.nu>

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the Licence, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston,
MA 02111-1307, USA.
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <ctype.h>
#include <math.h>
#include <unistd.h>
#include <sys/stat.h>
#include <time.h>

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/cursorfont.h>

#include <Mowitz/Mowitz.h>
#include <X11/xpm.h>
#include "MwHtmlP.h"

extern char *x_resolve_url(char *, char *);

static float floatOne = 1.0;

#define offset(field) XtOffsetOf(MwHtmlRec, html.field)
static XtResource resources[] = {
	{
		XtNurl,			/* name */
		XtCUrl,			/* class */
		XtRString,		/* type */
		sizeof(String),		/* size */
		offset(url),		/* offset */
		XtRImmediate,		/* default_type */
		(String)0		/* default_addr */
	}, {
		XtNtopRow,
		XtCTopRow,
		XtRInt,
		sizeof(int),
		offset(top_row),
		XtRImmediate,
		(XtPointer)0
	}, {
		XtNtopCol,
		XtCTopCol,
		XtRInt,
		sizeof(int),
		offset(top_col),
		XtRImmediate,
		(XtPointer)0
	}, {
		XtNtotalWidth,
		XtCTotalWidth,
		XtRInt,
		sizeof(int),
		offset(total_width),
		XtRImmediate,
		(XtPointer)0
	}, {
		XtNtotalHeight,
		XtCTotalHeight,
		XtRInt,
		sizeof(int),
		offset(total_height),
		XtRImmediate,
		(XtPointer)0
	}, {
		XtNzoom,
		XtCZoom,
		XtRFloat,
		sizeof(float),
		offset(zoom),
		XtRFloat,
		(XtPointer)&floatOne
	}, {
		XtNcallback,
		XtCCallback,
		XtRCallback,
		sizeof(XtPointer),
		offset(callbacks),
		XtRCallback,
		NULL
	}, {
		XtNchangeUrl,
		XtCChangeUrl,
		XtRCallback,
		sizeof(XtPointer),
		offset(change_url),
		XtRCallback,
		NULL
	}, {
		XtNstatus,
		XtCStatus,
		XtRWidget,
		sizeof(Widget),
		offset(status),
		XtRImmediate,
		(XtPointer)None,
	}, {
		XtNdelay,
		XtCDelay,
		XtRInt,
		sizeof(int),
		offset(delay),
		XtRImmediate,
		(XtPointer)0
	}
};
#undef offset

/* methods */
static void DoLayout(MwHtmlWidget);
static void Resize(Widget);
static XtGeometryResult GeometryManager(Widget,
		XtWidgetGeometry *, XtWidgetGeometry *);
static void ChangeManaged(Widget);
static void Redisplay(Widget, XEvent *, Region);
static void Initialize(Widget, Widget, ArgList, Cardinal *);
static void Realize(Widget, XtValueMask *, XSetWindowAttributes *);
static void Destroy(Widget);
static Boolean SetValues(Widget, Widget, Widget, ArgList, Cardinal *);

/* actions */

static void HtmlAction(Widget, XEvent *, String *, Cardinal *);
static void HtmlMotion(Widget, XEvent *, String *, Cardinal *);
static void HtmlButton(Widget, XEvent *, String *, Cardinal *);

static XtActionsRec actions[] =
{
	{"HtmlAction",		HtmlAction},
	{"HtmlMotion",		HtmlMotion},
	{"HtmlButton",		HtmlButton},
};

/* translations */
static char translations[] =
"<Key>:		HtmlAction() \n"
"<Motion>:	HtmlMotion() \n"
"<Btn1Down>:	HtmlButton() \n"
;

MwHtmlClassRec mwHtmlClassRec = {
  { /* core fields */
    /* superclass		*/	(WidgetClass) &compositeClassRec,
    /* class_name		*/	"MwHtml",
    /* widget_size		*/	sizeof(MwHtmlRec),
    /* class_initialize		*/	NULL,
    /* class_part_initialize	*/	NULL,
    /* class_inited		*/	FALSE,
    /* initialize		*/	Initialize,
    /* initialize_hook		*/	NULL,
    /* realize			*/	Realize,
    /* actions			*/	actions,
    /* num_actions		*/	XtNumber(actions),
    /* resources		*/	resources,
    /* num_resources		*/	XtNumber(resources),
    /* xrm_class		*/	NULLQUARK,
    /* compress_motion		*/	TRUE,
    /* compress_exposure	*/	TRUE,
    /* compress_enterleave	*/	TRUE,
    /* visible_interest		*/	FALSE,
    /* destroy			*/	Destroy,
    /* resize			*/	Resize,
    /* expose			*/	Redisplay,
    /* set_values		*/	SetValues,
    /* set_values_hook		*/	NULL,
    /* set_values_almost	*/	XtInheritSetValuesAlmost,
    /* get_values_hook		*/	NULL,
    /* accept_focus		*/	NULL,
    /* version			*/	XtVersion,
    /* callback_private		*/	NULL,
    /* tm_table			*/	translations,
    /* query_geometry		*/	XtInheritQueryGeometry,
    /* display_accelerator	*/	XtInheritDisplayAccelerator,
    /* extension		*/	NULL
  },{
/* composite_class fields */
    /* geometry_manager   */    GeometryManager,
    /* change_managed     */    ChangeManaged,
    /* insert_child       */    XtInheritInsertChild,
    /* delete_child       */    XtInheritDeleteChild,
    /* extension          */    NULL
  }, { /* html fields */
    /* empty			*/	0
  }
};

WidgetClass mwHtmlWidgetClass = (WidgetClass)&mwHtmlClassRec;

static void HtmlAction(Widget w, XEvent *event, String *params, Cardinal *n)
{
/*printf("HtmlAction(%s)\n", XtName(w))*/;
}

static object_box *locate_object(object_box *ob,
		int x_off, int y_off, int x, int y)
{
	object_box *ob1;
	if (ob == NULL) return NULL;

	if (ob->child == NULL &&
	    x > x_off+ob->x &&
	    x < x_off+ob->x+ob->w &&
	    y > y_off+ob->y &&
	    y < y_off+ob->y+ob->h) return ob;

	ob1 = locate_object(ob->child, x_off+ob->x, y_off+ob->y, x, y);
	if (ob1) return ob1;
	return locate_object(ob->next, x_off, y_off, x, y);
}

/* check if an object lives between ob1 and ob2 */
static int object_between(object_box *ob, object_box *ob1, object_box *ob2)
{
	while (ob1) {
		if (ob1 == ob) break;
		ob1 = ob1->next;
	}
	while (ob1) {
		if (ob1 == ob2) return 1;
		ob1 = ob1->next;
	}
	return 0;
}

/*
The links can look like this:

1:xxxxxxxxxLLLLLL
2:LLLLxxxxxxxxxxx

min: (x > x1 && y > y1) || (y > (y1+h1))
max: (x < x2 && y < (y2+h2)) || (y < y2)
*/
static char *get_href(MwHtmlWidget hw, int x, int y)
{
	int i;

	object_box *o = locate_object(hw->html.ob,
		hw->html.top_col, hw->html.top_row,
		x, y);

	if (o == NULL) {
		return NULL;
	}

	for (i = 0; i < hw->html.nref; i++) {
		if (object_between(o, hw->html.ref[i].ob1, hw->html.ref[i].ob2))
			return hw->html.ref[i].url;
	}
	return NULL;
}

/* to contains the current url which is overwritten with the new one,
   which is calculated from to and from.
 
   combine_url(<anything>, "http://host/") => "http://host/"
   combine_url("http://host/index.html", "demo/index.html")
        => "http://host/demo/index.html"
*/
static void combine_url(char *to, char *from)
{
	char *p;
 
	p = x_resolve_url(to, from);
	strcpy(to, p);
	MwFree(p);
}

static void dump_tree(object_box *ob, int level, int x_off, int y_off)
{
	int i;
	if (ob == NULL) return;
	for (i = 0; i < level; i++) printf("   ");
	printf(">%d@%d,%d<\n", ob->type, x_off+ob->x, y_off+ob->y);
	dump_tree(ob->child, level+1, x_off+ob->x, y_off+ob->y);
	dump_tree(ob->next, level, x_off, y_off);
}

static void HtmlMotion(Widget w, XEvent *event, String *params, Cardinal *n)
{
	char b[1024];
	MwHtmlWidget hw = (MwHtmlWidget)w;
	int x = event->xmotion.x+hw->html.top_col;
	int y = event->xmotion.y+hw->html.top_row;
	char *u = get_href((MwHtmlWidget)w, x, y);

	if (hw->html.url) strcpy(b, hw->html.url);
	else b[0] = '\0';
	if (u) {
		combine_url(b, u);
		XDefineCursor(XtDisplay(w), XtWindow(w), hw->html.cursor);
	} else {
		b[0] = '\0';
		XDefineCursor(XtDisplay(w), XtWindow(w), None);
	}
	if (hw->html.status) {
		MwLabelSet(hw->html.status, b);
	}
}

static void HtmlButton(Widget w, XEvent *event, String *params, Cardinal *n)
{
	char b[1024];
	MwHtmlWidget hw = (MwHtmlWidget)w;
	int x = event->xbutton.x+hw->html.top_col;
	int y = event->xbutton.y+hw->html.top_row;
	char *u = get_href((MwHtmlWidget)w, x, y);
	if (u) {
		if (hw->html.url) strcpy(b, hw->html.url);
		else b[0] = '\0';
		if (u) combine_url(b, u);
		else b[0] = '\0';
		XtCallCallbacks(w, XtNcallback, (XtPointer)b);
	}
}

static GC get_gc(Widget w, unsigned long fg, unsigned long bg, Font font)
{
        unsigned long valuemask = 0;
        XGCValues values;
        GC gc = XCreateGC(XtDisplay(w), XtWindow(w), valuemask, &values);

        XSetForeground(XtDisplay(w), gc, fg);
        XSetBackground(XtDisplay(w), gc, bg);
        if (font != -1)
                XSetFont(XtDisplay(w), gc, font);
        return gc;
}

#define superclass (&coreClassRec)
static void Realize(Widget w, XtValueMask *valueMask,
		XSetWindowAttributes *attributes)
{
	MwHtmlWidget rtw = (MwHtmlWidget)w;
	unsigned long fg, bg, blockbg;
	XColor screen_color, exact_color;
	Display *dpy = XtDisplay(w);
	char *name, *class;
	XIMStyles *styles;
	int i;

	(*superclass->core_class.realize)(w, valueMask, attributes);
	fg = BlackPixelOfScreen(XtScreen(w));
	bg = rtw->core.background_pixel;
	XAllocNamedColor(dpy,
		DefaultColormap(dpy, DefaultScreen(dpy)),
		"grey", &screen_color, &exact_color);
	blockbg = screen_color.pixel;
	rtw->html.clear_gc = get_gc(w, bg, fg, -1);
	rtw->html.cell_gc = get_gc(w, fg, blockbg,
			-1/*get_font(0)*/);

#ifdef HAVE_XCREATEIC
	/* Set up input methods */
	XtGetApplicationNameAndClass(dpy, &name, &class);
	rtw->html.xim = XOpenIM(dpy, XtDatabase(dpy), name, class);
	XGetIMValues(rtw->html.xim,
		XNQueryInputStyle, &styles,
		(char *)0);
	for (i = 0; i < styles->count_styles; i++) {
		if (styles->supported_styles[i] ==
		    (XIMPreeditNothing | XIMStatusNothing)) break;
	}
	if (i == styles->count_styles) i = 0;
	rtw->html.xic = XCreateIC(rtw->html.xim,
		XNInputStyle, styles->supported_styles[i],
		XNClientWindow, XtWindow(w),
		(char *)0);
#endif
}

static void Initialize(Widget request, Widget new, ArgList args, Cardinal *n)
{
	MwHtmlWidget rtw = (MwHtmlWidget)new;

	rtw->html.timeout = None;
	rtw->html.ref = NULL;
	rtw->html.nref = 0;
	rtw->html.ob = NULL;
	if (rtw->core.width == 0) rtw->core.width = 200;
	if (rtw->core.height == 0) rtw->core.height = 200;
	rtw->html.cursor = XCreateFontCursor(XtDisplay(new), XC_hand2);
	img_read("fish.png");
	rtw->html.fish = img_pop();
	if (rtw->html.fish == NULL) printf("fish failed\n");
	else printf("fish successful, size = %dx%d\n",
		rtw->html.fish->width, rtw->html.fish->height);
}

static void free_tree(MwHtmlNode *t)
{
	if (t) {
		free_tree(t->child);
		free_tree(t->next);
		MwFree(t);
	}
}

static void free_text(MwHtmlWidget hw)
{
	int i;
	/* free the hyperlinks as well, because they will be rebuilt */
	for (i = 0; i < hw->html.nref; i++) {
		MwFree(hw->html.ref[i].url);
	}
	MwFree(hw->html.ref);
	hw->html.ref = NULL;
	hw->html.nref = 0;
#if 1
	MwHtmlFree(hw->html.ob);
	hw->html.ob = NULL;
#endif
}

static void Destroy(Widget w)
{
	MwHtmlWidget rtw = (MwHtmlWidget)w;

	XFreeGC(XtDisplay(w), rtw->html.clear_gc);
	XFreeGC(XtDisplay(w), rtw->html.cell_gc);
	img_free(rtw->html.fish);
#ifdef HAVE_XCREATEIC
	XDestroyIC(rtw->html.xic);
	XCloseIM(rtw->html.xim);
#endif
	if (rtw->html.timeout)
		XtRemoveTimeOut(rtw->html.timeout);
	free_text(rtw);
}

struct hs_out {
	float x, y/*, xmax*/;	/* where to put the next char */
	int il;			/* indentation level */
	object_box *fob;	/* first obj in current line */
	MwHtmlWidget w;
	GC g;
	float z;		/* zoom; currently unused (always 1.0 */
	href ref;
	int refstate;		/* 0 = outside, 1 = inside ref */
};

static void begin_a(void *closure, object_box *ob, MwHtmlAnchor *a)
{
	struct hs_out *o = closure;
	MwHtmlWidget hw = o->w;
	int nref = hw->html.nref;
	o->refstate = 0;
	if (a->href == NULL && a->name == NULL) return;
	hw->html.ref = MwRealloc(hw->html.ref, (nref+1)*sizeof(href));
	if (a->href) o->ref.url = MwStrdup(a->href);
	else o->ref.url = NULL;
	if (a->name) o->ref.name = MwStrdup(a->name);
	else o->ref.name = NULL;
	o->ref.ob1 = ob;
	o->refstate = 1;
}

static void end_a(void *closure, object_box *ob)
{
	struct hs_out *o = closure;
	MwHtmlWidget hw = o->w;
	int nref = hw->html.nref;
	if (o->refstate == 0) {
		fprintf(stderr, "Error: </a> not in ref\n");
		return;
	}
	hw->html.nref = nref+1;
	o->ref.ob2 = ob;
	hw->html.ref[nref] = o->ref;
}

static object_box *next_up(object_box *ob)
{
	if (ob == NULL) return NULL;
	if (ob->next) return ob->next;
	return next_up(ob->parent);
}

static float breakline(object_box *fob, /*struct hs_out *o,*/ object_box *ob)
{
	float w, mh = 0;
	object_box *ob1;

	/* eliminate spaces at the beginning of the line */
	ob1 = /*o->*/fob;
	w = 0;
	while (ob1 && ob1 != ob &&
			(ob1->w == 0 || ob1->type == MW_HTML_SPACE)) {
		w += ob1->w;
		ob1->w = 0;
		ob1 = ob1->next;
	}
	if (ob1 == NULL) return 0;
	while (ob1 != ob) {
		ob1->x -= w;
		ob1 = ob1->next;
	}

	for (ob1 = /*o->*/fob; ob1 != ob; ob1 = ob1->next) {
		if (ob1->type == MW_HTML_SPACE) {
			ob1->w = 0;
		}
	}

	for (ob1 = /*o->*/fob; ob1 != ob; ob1 = ob1->next) {
		if (ob1->h > mh) {
			mh = ob1->h;
		}
	}

	for (ob1 = /*o->*/fob; ob1 != ob; ob1 = ob1->next) {
		ob1->y = ob1->y + mh - ob1->h;
	}

	return mh;
}

static void draw_input(MwHtmlWidget hw, Drawable d,
		object_box *ob, int x, int y)
{
	MwHtmlInput *a = ob->data;
	switch (a->type) {
	default:
printf("draw input type %d at %d,%d\n", a->type, x, y);
	}
}

static void draw_pic(MwHtmlWidget hw, Drawable d, image *img, int x, int y)
{
	int i, j;
	XImage *im_out;
	GC gc = hw->html.cell_gc;
	Display *dpy = XtDisplay((Widget)hw);
	int screen = DefaultScreen(dpy);
	Visual *visual = DefaultVisual(dpy, screen);
	int bitmap_pad = XBitmapPad(dpy);
	int format = ZPixmap;
	int offset = 0;
	char *data = NULL;
	int width, height, depth;
	int bytes_per_line = 0;
	XColor color;
	pixel p;

	if (img == NULL) return;
	width = img->width;
	height = img->height;
	depth = hw->core.depth;
	im_out = XCreateImage(dpy, visual, depth, format, offset, data,
			width, height, bitmap_pad, bytes_per_line);
	im_out->data = MwMalloc(im_out->bytes_per_line*im_out->height);
	for (i = 0; i < height; i++) {
		for (j = 0; j < width; j++) {
			p = GET_PIXEL(img, j, i);
			color.red = 257*p.r;
			color.green = 257*p.g;
			color.blue = 257*p.b;
			MwAllocColor(dpy, None, &color);
			XPutPixel(im_out, j, i, color.pixel);
		}
	}
	XPutImage(dpy, d, gc, im_out, 0, 0,
		x, y+1, img->width, img->height);
	XDestroyImage(im_out);
}

static image *load_pic(char *base, char *url)
{
	char *nurl;
	image *img;

	nurl = x_resolve_url(base, url);
	img = img_load(nurl);
	MwFree(nurl);
	return img;
}

static void size_objects(MwHtmlWidget rtw, struct hs_out *o, object_box *ob)
{
	int table_level, table_rows, table_cols, table_max_cols;
	object_box *table_start, *table_end, *ob1, *ob2;
	image *img;
	MwRichchar *rc;
	MwRichchar space[2] = {{' ', 0}, {'\0', 0}};
	MwHtmlRow *row_a;
	MwHtmlTable *table_a;
	int col;

	table_level = table_rows = table_cols = table_max_cols = 0;
	table_start = table_end = NULL;

	if (ob == NULL) {
		return;	/* nothing to do */
	}

	/* calculate size of all objects */

	for (/*ob = rtw->html.ob*/; ob; ob = ob->next) {
		ob->h = 0;
		ob->min_w = ob->max_w = 0;
		switch (ob->type) {
		case MW_HTML_BEGIN_A:
			begin_a(o, ob, ob->data);
			break;
		case MW_HTML_END_A:
			end_a(o, ob);
			break;
		case MW_HTML_IMAGE:
			img = load_pic(rtw->html.url,
				((MwHtmlImage *)ob->data)->src);
			if (img) {
				ob->min_w = ob->max_w = img->width;
				ob->h = img->height;
				img_free(img);
			} else if (rtw->html.fish) {
				ob->min_w = ob->max_w = rtw->html.fish->width;
				ob->h = rtw->html.fish->height;
			}

			break;
		case MW_HTML_INPUT:
			ob->min_w = ob->max_w = 100;	/* what do I know */
			ob->h = 30;
			break;
		case MW_HTML_INDENT:
			o->il = (int)ob->data;
			ob->min_w = ob->max_w = 0;
			break;
		case MW_HTML_HR:
			ob->min_w = ob->max_w = rtw->core.width;
			ob->h = 1;
			break;
		case MW_HTML_WORDPART:
			rc = ob->data;
			ob->h = MwRcStrheight(rc, -1);
			ob->min_w = ob->max_w = MwRcStrwidth(rc, -1);
			break;
		case MW_HTML_SPACE:
			rc = ob->data;
			space[0].fmt = rc[0].fmt;
			ob->h = MwRcStrheight(space, -1);
			ob->min_w = ob->max_w = MwRcStrwidth(space, -1);
			break;
		case MW_HTML_CELL:
			size_objects(rtw, o, ob->child);
			for (ob1 = ob->child; ob1; ob1 = ob1->next) {
				if (ob1->min_w > ob->min_w)
					ob->min_w = ob1->min_w;
				ob->max_w += ob1->max_w;
			}
			break;
		case MW_HTML_ROW:
			row_a = ob->data = MwMalloc(sizeof *row_a);
			row_a->nchild = 0;
			size_objects(rtw, o, ob->child);
			for (ob1 = ob->child; ob1; ob1 = ob1->next) {
				row_a->nchild++;
			}
			break;
		case MW_HTML_TABLE:
			table_a = ob->data = MwMalloc(sizeof *table_a);
			size_objects(rtw, o, ob->child);
			table_a->max_child = 0;
			for (ob1 = ob->child; ob1; ob1 = ob1->next) {
				row_a = ob1->data;
				if (row_a->nchild > table_a->max_child) {
					table_a->max_child = row_a->nchild;
				}
			}
			table_a->colwidth = MwMalloc(table_a->max_child *
						sizeof *(table_a->colwidth));
			for (col = 0; col < table_a->max_child; col++) {
				table_a->colwidth[col].min = 0;
				table_a->colwidth[col].max = 0;
			}
			/* get largest min_w/max_w from each column */
			for (ob1 = ob->child; ob1; ob1 = ob1->next) {
				col = 0;
				for (ob2 = ob1->child; ob2; ob2 = ob2->next) {
					if (ob2->min_w > table_a->colwidth[col].min)
						table_a->colwidth[col].min = ob2->min_w;
					if (ob2->max_w > table_a->colwidth[col].max)
						table_a->colwidth[col].max = ob2->max_w;
					col++;
				}
			}
			for (col = 0; col < table_a->max_child; col++) {
				ob->min_w += table_a->colwidth[col].min;
				ob->max_w += table_a->colwidth[col].max;
			}
			break;
		default:
			/* workaround for the new parser */
			if (ob->child) {
printf("you are not allowed to have children\n");
exit(0);
				size_objects(rtw, o, ob->child);
			}
			break;
		}
	}
}

static void assign_sizes(MwHtmlWidget rtw, /*struct hs_out *o,*/ object_box *ob)
{
	MwHtmlTable *table_a; int col;
	object_box *fob, *ob1, *ob2;
	int x, y;
	int W, D, d;

	if (ob == NULL) {
		return;
	}

	for (; ob; ob = ob->next) {
		switch (ob->type) {
		case MW_HTML_TABLE:
			table_a = ob->data;
			ob->w = 0;
			if (ob->min_w > ob->parent->w) {
				for (col = 0; col < table_a->max_child; col++) {
					table_a->colwidth[col].real =
						table_a->colwidth[col].min;
				}
			} else if (ob->max_w < ob->parent->w) {
				for (col = 0; col < table_a->max_child; col++) {
					table_a->colwidth[col].real =
						table_a->colwidth[col].max;
				}
			} else {
				W = ob->parent->w - ob->min_w;
				D = ob->max_w - ob->min_w;
				if (D == 0) D = 1;	/* no div by zero */
				for (col = 0; col < table_a->max_child; col++) {
					d = table_a->colwidth[col].max -
						table_a->colwidth[col].min;
					table_a->colwidth[col].real =
						table_a->colwidth[col].min+d*W/D;
				}
			}
			for (col = 0; col < table_a->max_child; col++) {
				ob->w += table_a->colwidth[col].real;
			}
			for (ob1 = ob->child; ob1; ob1 = ob1->next) {
				ob1->w = ob->w; ob1->x = 0;	/* always */
				col = 0;
				x = 0;
				for (ob2 = ob1->child; ob2; ob2 = ob2->next) {
					ob2->w = table_a->colwidth[col].real;
					ob2->x = x;
					x += ob2->w;
					ob2->y = 0;     /* always */
					col++;
				}
			}
			assign_sizes(rtw, /*o,*/ ob->child);
			ob->h = 0;
			for (ob1 = ob->child; ob1; ob1 = ob1->next) {
				ob1->y = ob->h;
				ob->h += ob1->h;
			}
			break;
		case MW_HTML_ROW:
			/* widths were done for us by the table, so we
			   only need to calculate our height */
			assign_sizes(rtw, /*o,*/ ob->child); /* do cells first */
			ob->h = 0;
			for (ob1 = ob->child; ob1; ob1 = ob1->next) {
				if (ob1->h > ob->h) ob->h = ob1->h;
			}
			break;
		case MW_HTML_CELL:
			/* widths were done by the table, so we
			   only need to calculate our height */
			assign_sizes(rtw, /*o,*/ ob->child); /* can be tables! */
			x = y = 0;
			fob = ob->child;	/* first in line */
			for (ob1 = ob->child; ob1; ob1 = ob1->next) {
				if (ob1->type == MW_HTML_NEWLINE ||
				    x+ob1->w > ob->w) {
					y += breakline(fob, ob1);
					x = 0;	/* FIXME: should be indent */
					fob = ob1;
				}
				ob1->x = x;
				x += ob1->w;
				ob1->y = y;
			}
			if (x) y += breakline(fob, NULL);
			ob->h = y;
			break;
		default:
			ob->w = ob->min_w;
		}
	}
}

static void draw_objects(MwHtmlWidget rtw, Drawable d, int x_off, int y_off,
		/*struct hs_out *o,*/ object_box *ob)
{
	image *img;
	MwRichchar *rc;
	MwRichchar space[2] = {{' ', 0}, {'\0', 0}};
	char b[100];

	/* draw all relevant objects */
	for (; ob; ob = ob->next) {
		if (ob->y+ob->h+y_off < 0) continue;
		if (ob->y+y_off > rtw->core.height) continue;
		switch (ob->type) {
		case MW_HTML_IMAGE:
			img = load_pic(rtw->html.url,
				((MwHtmlImage *)ob->data)->src);
			if (img) {
				draw_pic(rtw, d, img,
					ob->x+x_off, ob->y+y_off);
				img_free(img);
			} else {
				draw_pic(rtw, d, rtw->html.fish,
					ob->x+x_off, ob->y+y_off);
			}

			break;
		case MW_HTML_INPUT:
			draw_input(rtw, d, ob, ob->x+x_off, ob->y+y_off);
			break;
		case MW_HTML_HR:
			break;
		case MW_HTML_WORDPART:
			rc = ob->data;
			MwRcStrdraw(d, rtw->html.cell_gc, ob->x, ob->y+ob->h,
				x_off, y_off, rc, -1, rtw->html.zoom);
			break;
		case MW_HTML_SPACE:
			rc = ob->data;
			space[0].fmt = rc[0].fmt;
			MwRcStrdraw(d, rtw->html.cell_gc, ob->x, ob->y+ob->h,
				x_off, y_off, space, -1, rtw->html.zoom);
			break;
		case MW_HTML_UBULLET:
			rc = MwRcMakerich("*", 0);
			MwRcStrdraw(d, rtw->html.cell_gc, ob->x-20, ob->y+ob->h,
				x_off, y_off, rc, -1, rtw->html.zoom);
			MwFree(rc);
			break;
		case MW_HTML_OBULLET:
			sprintf(b, "%d.", (int)ob->data);
			rc = MwRcMakerich(b, 0);
			MwRcStrdraw(d, rtw->html.cell_gc, ob->x-20, ob->y+ob->h,
				x_off, y_off, rc, -1, rtw->html.zoom);
			MwFree(rc);
			break;
		default:
			/* workaround for the new parser */
			if (ob->child) {
				draw_objects(rtw, d,
					x_off+ob->x, y_off+ob->y,
					ob->child);
			}
		}
	}
}

static void MwHtmlDraw(MwHtmlWidget rtw, Drawable d)
{
	struct hs_out o;

	if (rtw->html.ob == NULL) {
		rtw->html.ob = MwHtmlParse(rtw->html.url);

		o.il = 0;
		o.x = 0;
		o.y = 0;
		o.w = rtw;
		o.g = rtw->html.cell_gc;
		o.z = rtw->html.zoom;
		o.fob = rtw->html.ob;
		rtw->html.ob->w = rtw->core.width;
		rtw->html.ob->h = 0;	/* we don't care yet */
		rtw->html.ob->x = rtw->html.ob->y = 0;
		size_objects(rtw, &o, rtw->html.ob);
		assign_sizes(rtw, /*&o,*/ rtw->html.ob);
		rtw->html.total_width = rtw->html.ob->w;
		rtw->html.total_height = rtw->html.ob->h;

		XtCallCallbacks((Widget)rtw, XtNchangeUrl,
			(XtPointer)rtw->html.url);
	}

dump_tree(rtw->html.ob, 1, rtw->html.top_col, rtw->html.top_row);

	draw_objects(rtw, d, -rtw->html.top_col, -rtw->html.top_row,
			/*&o,*/ rtw->html.ob);
}

/* ---
Draw onto a pixmap.
*/

static Pixmap html_pixmap(MwHtmlWidget rtw)
{
        int width, height, depth;
        Pixmap scribble;
        Widget w = (Widget)rtw;

	width = rtw->core.width;
	height = rtw->core.height;
	depth = rtw->core.depth;
	if (width > 2000 || height > 2000) return None;

        scribble = XCreatePixmap(XtDisplay(w), XtWindow(w),
                width, height, depth);
        XFillRectangle(XtDisplay(w), scribble, rtw->html.clear_gc,
                0, 0, width, height);
        MwHtmlDraw(rtw, scribble);
        return scribble;
}

static void do_redisplay(XtPointer client_data, XtIntervalId *id)
{
	Widget w = (Widget)client_data;
        Pixmap scribble;
        GC gc;
        unsigned long valuemask = 0;
        XGCValues values;
	MwHtmlWidget rtw = (MwHtmlWidget)w;

	scribble = html_pixmap(rtw);
        if (scribble == None) return;
        gc = XCreateGC(XtDisplay(w), XtWindow(w),
                        valuemask, &values);
        XCopyArea(XtDisplay(w), scribble, XtWindow(w),
                gc, 0, 0, rtw->core.width, rtw->core.height, 0, 0);
        XFreePixmap(XtDisplay(w), scribble);
        XFreeGC(XtDisplay(w), gc);
	/* update plugin positions */
	DoLayout(rtw);
	
	/* clear the timeout */
	rtw->html.timeout = None;
}

static void Redisplay(Widget w, XEvent *event, Region r)
{
	MwHtmlWidget tw = (MwHtmlWidget)w;
	if (tw->html.timeout) return;	/* already set */
	if (tw->html.delay) {
		tw->html.timeout = XtAppAddTimeOut(
			XtWidgetToApplicationContext(w),
			tw->html.delay, do_redisplay, (XtPointer)w);
	} else {
		do_redisplay((XtPointer)w, NULL);
	}
}

static Boolean SetValues(Widget current, Widget request, Widget new,
		ArgList args, Cardinal *nargs)
{
	MwHtmlWidget hw = (MwHtmlWidget)new;
	MwHtmlWidget cw = (MwHtmlWidget)current;
	Boolean do_redisplay = True;

	if (hw->html.url != NULL && hw->html.url != cw->html.url) {
		hw->html.url = x_resolve_url(cw->html.url, hw->html.url);
		MwFree(cw->html.url);
		free_text(hw);
	}
	if (hw->html.ob == NULL || hw->html.ob != cw->html.ob) {
		free_text(hw);
	}
	if (do_redisplay) Redisplay(new, NULL, None);
	return False;
}

/*
 * Do a layout, actually assigning positions.
 *
 * This function uses a callback from the application to get the
 * position of each plugin. In the case of PW, that callback uses
 * richtext_char2coords. See pw/window.c.
 */

static void DoLayout(MwHtmlWidget sw)
{
/*printf("Don't know what to do with children\n")*/;
}

/*
 * Actually layout the table
 */

static void Resize(Widget w)
{
	MwHtmlWidget hw = (MwHtmlWidget)w;
	free_text(hw);
        DoLayout(hw);
}

/*
 * Geometry Manager
 */

static XtGeometryResult GeometryManager(Widget w,
                XtWidgetGeometry *request, XtWidgetGeometry *reply)
{
        return XtGeometryYes;
}

static void ChangeManaged(Widget w)
{
    DoLayout((MwHtmlWidget)w);
}

int MwHtmlLookupString(Widget w, XEvent *event, char *buf,
			 int bufsiz, KeySym *keysym)
{
#ifdef HAVE_XCREATEIC
	Status status;
	return XmbLookupString(((MwHtmlWidget)w)->html.xic,
				(XKeyEvent *)event, (char *)buf, bufsiz,
				keysym, &status);
#else
        return XLookupString((XKeyEvent *)event, (char *)buf, bufsiz,
				keysym, NULL);
#endif
}

void MwHtmlSetZoom(Widget w, float zoom)
{
	MwHtmlWidget rw = (MwHtmlWidget)w;

	if (zoom < .1) zoom = .1;
	if (zoom > 10) zoom = 10;
	if (rw->html.zoom != zoom) {
		rw->html.zoom = zoom;
		Redisplay(w, NULL, None);
	}
}

